"""Command line interface"""
import os
import os.path
import time
import buildver
from buildver import error


__all__ = []


HELP = """\
Build and versioning tool for Python projects.

Check a project
===============
$ buildver check

Set a new version 
=================
$ buildver set <version>

Build your project 
==================
$ buildver build

Build then set a new version
============================
$ buildver build then <version>


Note: Use "+maj", "+min" and "+rev" to increment
the current version of your project. These version
modifiers increment the "major", "minor", and "revision"
numbers respectively.

 Current | Modifier | Next   |
---------|----------|--------|
 1.0.0   | +rev     | 1.0.1  |
 1.0.1   | +min     | 1.1.0  |
 1.1.0   | +maj     | 2.0.0  |
 
By default, the command "build" will build a distribution
package, then automatically increment the revision number
of the current version. If you don't want the version number
incremented automatically, use the "pass" modifier:
$ buildver build then pass

Read more online:
https://github.com/pyrustic/buildver
"""


def print_help(*args, project_dir=None):
    print(HELP)


def check_project(*args, project_dir=None):
    project_dir = project_dir if project_dir else os.getcwd()
    try:
        project_version = buildver.get_version(project_dir)
    except error.MissingVersionFileError as e:
        print("Missing VERSION file.")
        return
    else:
        if not project_version:
            project_version = "0.0.0"
            buildver.set_version(project_version, project_dir)
    project_name = os.path.basename(project_dir)
    print("{} v{} (source)".format(project_name, project_version))
    # Latest build
    latest_build_info = buildver.get_latest_build(project_dir)
    if not latest_build_info:
        return
    version, timestamp = latest_build_info
    time_elapsed = _stringify_time_elapsed(timestamp)
    time_elapsed = "{}".format(time_elapsed) if time_elapsed else ""
    print(".whl v{} (package) built {}".format(version, time_elapsed))


def set_version(*args, project_dir=None):
    if not args:
        print("Missing version number.")
        return
    if len(args) > 1:
        print("Wrong usage of the command.")
        return
    project_dir = project_dir if project_dir else os.getcwd()
    new_version = args[0]
    if " " in new_version:
        print("Invalid version.")
        return
    current_version = None
    try:
        current_version = buildver.get_version(project_dir)
    except error.MissingVersionFileError as e:
        pass
    current_version = current_version if current_version else "0.0.0"
    if new_version:
        try:
            new_version = buildver.interpret_version(current_version, new_version)
        except ValueError as e:
            print("Failed to set the version. Please check the VERSION file.")
            return
        except Exception as e:
            print("Failed to set the version.")
            return
    else:
        new_version = "0.0.1"
    if current_version == new_version:
        print("VERSION file not updated")
    else:
        buildver.set_version(new_version, project_dir)
        print("VERSION file updated from {} to {}".format(current_version,
                                                          new_version))


def build_project(*args, project_dir=None):
    if not args:
        next_version = "+rev"
    elif len(args) == 1:
        print("Wrong usage of the command.")
        return
    else:
        if args[0] == "then":
            next_version = args[1]
        else:
            print("Wrong usage of the command.")
            return
    project_dir = project_dir if project_dir else os.getcwd()
    try:
        cur_version = buildver.get_version(project_dir)
    except error.MissingVersionFileError as e:
        print("Missing VERSION file.")
        return
    # BUILD
    print("building v{} ...".format(cur_version))
    success, _ = buildver.build_project(project_dir)
    if success:
        buildver.update_build_report(cur_version, project_dir)
    else:
        print("Failed to build a distribution package")
        return
    project_name = os.path.basename(project_dir)
    print("Successfully built '{}' v{} !".format(project_name, cur_version))
    # set next version
    set_version(next_version, project_dir=project_dir)


def _stringify_time_elapsed(timestamp):
    if not timestamp:
        return ""
    timestamp = int(timestamp)
    now = time.time()
    seconds_elapsed = int(now - timestamp)
    years, days, hours, minutes, seconds = _time_elapsed(seconds_elapsed)
    if years:
        return _x_ago(years, "year")
    if days:
        return _x_ago(days, "day")
    if hours:
        return _x_ago(hours, "hour")
    if minutes:
        return _x_ago(minutes, "min")
    if seconds:
        return _x_ago(seconds, "sec")


def _x_ago(x, name):
    plural = ""
    if x > 1:
        plural = "s"
    return "{} {}{} ago".format(x, name, plural)


def _time_elapsed(seconds_elapsed):
    s_in_min = 60
    s_in_hour = s_in_min * 60
    s_in_day = s_in_hour * 24
    s_in_year = s_in_day * 365
    years, remainder = divmod(seconds_elapsed, s_in_year)
    days, remainder = divmod(seconds_elapsed, s_in_day)
    hours, remainder = divmod(remainder, s_in_hour)
    minutes, seconds = divmod(remainder, s_in_min)
    return years, days, hours, minutes, seconds



import os
import os.path
import time
import subrun
from buildver import error


def get_version(project_dir=None):
    """
    This function read the VERSION file in the project_dir project
    then returns the version (str) of the project.

    [parameters]
    - project_dir: str, path to the project_dir project (default to os.getcwd)

    [exceptions]
    - error.MissingVersionFileError: raised when the VERSION file is missing

    [return]
    str, version extracted from $PROJECT_DIR/VERSION or None
    """
    if not project_dir:
        project_dir = os.getcwd()
    path = os.path.join(project_dir, "VERSION")
    if not os.path.exists(path):
        raise error.MissingVersionFileError
    with open(path, "r") as file:
        lines = file.readlines()
    if not lines:
        return None
    line = lines[0]
    cache = []
    for char in line:
        if char not in (" ", "\n"):
            cache.append(char)
    version = "".join(cache)
    version = None if (not version or version.isspace()) else version
    return version


def set_version(version, project_dir=None):
    """
    This function edits the content of $PROJECT_DIR/VERSION

    [parameters]
    - version: str, the version
    - project_dir: str, path to the project_dir project (default: os.getcwd)

    [exceptions]
    - item: val

    [return]
    Returns False if the project_dir doesn't exist, returns True if all right
    """
    if not project_dir:
        project_dir = os.getcwd()
    if not os.path.exists(project_dir):
        return False
    path = os.path.join(project_dir, "VERSION")
    with open(path, "w") as file:
        file.write(version)
    return True


def interpret_version(cur_version, new_version):
    """
    This function interprets the command to set a new version.

    [parameters]
    - cur_version: str, the current version, the one to alter.
    - new_version: str, the new version. It can be a canonical version like "0.0.1"
    or a version modifier like "+rev" or "pass".
    Use "+maj" to increment the major number of the current version,
    "+min" to increment the minor number of the current version,
    and "+rev": to increment the revision number of the current version.

    [return]
    The new version as it should be saved in $PROJECT_DIR/VERSION
    """
    if new_version == "pass":
        return cur_version
    if new_version not in ("+maj", "+min", "+rev"):
        return new_version
    cache = cur_version.split(".")
    cache_size = len(cache)
    if cache_size == 1:
        cache.extend(["0", "0"])
    # interpret '+maj', '+min' and '+rev'
    if new_version == "+maj":
        cache[0] = _increment(cache[0])
        _reset_to_zero(cache, 1, cache_size)
    elif new_version == "+min":
        cache[1] = _increment(cache[1])
        _reset_to_zero(cache, 2, cache_size)
    elif new_version == "+rev":
        cache[-1] = _increment(cache[-1])
    version = ".".join(cache)
    return version


def build_project(project_dir=None):
    """
    Build the project

    [parameters]
    - project_dir: str, project dir (default: os.getcwd)

    [return]
    Returns a tuple: success_boolean, error_str
        """
    if not project_dir:
        project_dir = os.getcwd()
    command = "python -m setup --quiet sdist bdist_wheel"
    info = subrun.ghostrun(command, cwd=project_dir)
    success = True
    if info.return_code != 0:
        success = False
    return success, info.error


def get_latest_build(project_dir=None):
    """
    Get basic information about the latest build

    [parameters]
    - project_dir: str, project dir (default: os.getcwd)

    [return]
    Returns a tuple of strings: (version, timestamp).
    Return None if there is nothing to show.
    """
    if not project_dir:
        project_dir = os.getcwd()
    build_report_path = os.path.join(project_dir, ".pyrustic",
                                     "buildver", "build_report")
    if not os.path.isfile(build_report_path):
        return None
    with open(build_report_path, "r") as file:
        line = file.readline()
        if not line:
            return None
    version, timestamp = line.split()
    return version, timestamp


def update_build_report(version, project_dir=None):
    """
    Update the build_report file

    [parameters]
    - version: str, version at its canonical form
    - project_dir: str, project dir (default: os.getcwd)

    [return]
    None
    """
    if not project_dir:
        project_dir = os.getcwd()
    build_report_dir = os.path.join(project_dir, ".pyrustic",
                                    "buildver")
    if not os.path.isdir(build_report_dir):
        os.makedirs(build_report_dir)
    build_report_path = os.path.join(build_report_dir, "build_report")
    if not os.path.isfile(build_report_path):
        with open(build_report_path, "w") as file:
            pass
    text = "{} {}".format(version, int(time.time()))
    # readlines
    with open(build_report_path, "r") as file:
        lines = file.readlines()
        lines.insert(0, "{}\n".format(text))
    # update build report file
    with open(build_report_path, "w") as file:
        data = "".join(lines[0:999])
        file.write(data)


def _increment(number):
    number = int(number) + 1
    return str(number)


def _reset_to_zero(target, from_index, to_index):
    for i in range(from_index, to_index):
        target[i] = "0"



import sys
from buildver import cli


__all__ = []


COMMANDS = {"help": cli.print_help,
            "check": cli.check_project,
            "set": cli.set_version,
            "build": cli.build_project}


def main():
    args = sys.argv[1:]
    # no args
    if not args:
        cli.print_help()
        return
    command = args[0].lower()
    arguments = args[1:]
    handler = COMMANDS.get(command)
    if handler:
        handler(*arguments)
    else:
        valid_commands = ", ".join(["'{}'".format(key) for key in COMMANDS.keys()])
        print("Valid commands are: {}.".format(valid_commands))


if __name__ == "__main__":
    main()
